# =====================================================================
# csa-client.awk: CSA-RPC response parser.
#
# Copyright (c) 2003,2007 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================

BEGIN {

  RS = "<"

  _FORMAT = "CALL"				# set default
}

# detect <rdf:RDF ...>
/^rdf:RDF[ \t>]/				{found=level=1; next}

# detect <... rdf:about="#call">
found && level==1 && /^[-a-zA-Z0-9]+[\t ]+rdf:about="#call">/ {

  rdf_element = $0
  sub(/[\t ]+rdf:about="#call">.*/,"",rdf_element)
  rdf_element = "^/" _escreg(rdf_element) ">"

  level++
  next
}

# detect <param>
found && level==2 && /^csa:param>/		{level++; next}

# Extract call arguments.
found && level==3 && /^rdf:Seq[\t ]+rdf:about="#call\./ {

  sub(/^.*rdf:about="#call\./,""); sub(/".*/,"")

  # replace invalid characters with underscores (_) in name.
  gsub(/[^A-Za-z0-9_]/, "_")

  # always skip call.0 in call arguments.
  if ($0 != "0") current_name = $0

  level++
  next
}

# detect <... rdf:about="#response">
found && level==1 && /^[-a-zA-Z0-9]+[\t ]+rdf:about="#response">/ {

  rdf_element = $0
  sub(/[\t ]+rdf:about="#response">.*/,"",rdf_element)
  rdf_element = "^/" _escreg(rdf_element) ">"

  level++
  next
}

# detect <result> or <error>, and extract result name.
found && level==2 && /^csa:(result|error)>/ {

  if (/^csa:error>/) _FORMAT = "ERROR"
  else _FORMAT = "RESULT"

  level++
  next
}

# detect <rdf:Seq rdf:about="#response....">
found && level==3 && /^rdf:Seq[ \t]+rdf:about="#response\.[1-9][0-9]*/ {

  level++
  next
}

# detect <rdf:li[^=]+rdf:datatype="...> in result and extract datatype
found && level==4 && current_name == "" && /^rdf:li[^=]+rdf:datatype="/ {

  # Support three common forms of schema reference.
  if (sub(/^[^>]*XMLSchema#/,""));
  else if (sub(/^[^>]*&xsd;/,""));
  else sub(/^[^>]*xsd:/,"")

  rdf_data[++rdf_num]=$0; sub(/".*/,"")

  # See http://www.w3.org/TR/2001/REC-xmlschema-2-20010502/#built-in-datatypes
  # for more info. Unknown types default to "string".

  if ($0 !~ /^(int|base64(Binary)?|boolean|double)$/)
					rdf_type[rdf_num]="string"
  else rdf_type[rdf_num] = $0

  rdf_type[0]++

  # detect empty tag.
  if (rdf_data[rdf_num] ~ /[^>]*\/>/) rdf_data[rdf_num] = ""
  else {
     sub(/[^>]*>/,"",rdf_data[rdf_num])

     if (rdf_type[rdf_num] ~ /^(int|double)$/) rdf_data[rdf_num] += 0

     else if (rdf_type[rdf_num] == "base64Binary") {
	# I use the extra qualifier "-csaRpc.", not to overwrite an
	# equally-named file uploaded by the client in the same CSA run.
	rdf_data[rdf_num] = ENVIRON["TMPDIR"] \
		"/hexBinary." _PID "." current_name "-csaRpc.tmp"
	if (_CMD != "") close(_CMD)		# just in case
	_CMD = "exec swu-encdec -d -b > " rdf_data[rdf_num]
	printf("%s", $0) | _CMD
	close(_CMD)
     }

     else rdf_data[rdf_num] = _xmldecode(rdf_data[rdf_num])
  }

  rdf_data[0]++

  # add type-specific and misc. checks on rdf_data[rdf_num] here ...

  # Note: large embedded data objects should never be inserted in
  # the response stream, but should rather be pointed to with either
  # a respone URI or a Content-Location HTTP response header, according
  # to REST.
}

# detect <rdf:li rdf:resource=".../> in result and extract the associated URI
found && level==4 && current_name == "" && /^rdf:li[^=]+rdf:resource="/ {

  sub(/[^=]+resource="/,""); sub(/".*/,"")

  rdf_type[++rdf_num] = "anyURI"
  rdf_type[0]++

  # Local file resource (file:) can be dangerous if trusted blindly by
  # the receiving application. The latter will have to make sure that
  # they are interpreted as file-paths only if they are also referenced
  # by the _ATTACHMENT[] array.

  rdf_data[rdf_num] = $0
  rdf_data[0]++

  # add URI-specific and misc. checks on rdf_data[rdf_num] here ...

}

# detect <rdf:li[^=]+rdf:datatype="...> in request and extract datatype
found && level==4 && current_name != "" && /^rdf:li[^=]+rdf:datatype="/ {

  # Support three common forms of schema reference.
  if (sub(/^[^>]*XMLSchema#/,""));
  else if (sub(/^[^>]*&xsd;/,""));
  else sub(/^[^>]*xsd:/,"")

  rdf_cdat[++rdf_cnum]=$0; sub(/".*/,"")

  # See http://www.w3.org/TR/2001/REC-xmlschema-2-20010502/#built-in-datatypes
  # for more info. Unknown types default to "string".

  if ($0 !~ /^(int|base64(Binary)?|boolean|double)$/)
					rdf_ctyp[rdf_cnum]="string"
  else rdf_ctyp[rdf_cnum] = $0

  rdf_ctyp[0]++

  rdf_cnam[rdf_cnum] = current_name; rdf_cnam[0]++

  # detect empty tag.
  if (rdf_cdat[rdf_cnum] ~ /[^>]*\/>/) rdf_cdat[rdf_cnum] = ""
  else {
     sub(/[^>]*>/,"",rdf_cdat[rdf_cnum])
     if (rdf_ctyp[rdf_cnum] ~ /^(int|double)$/) rdf_cdat[rdf_cnum] += 0
     else rdf_cdat[rdf_cnum] = _xmldecode(rdf_cdat[rdf_cnum])
  }

  rdf_cdat[0]++

  # add type-specific and misc. checks on rdf_cdat[rdf_cnum] here ...
}

# detect <rdf:li rdf:resource=".../> in request and extract the associated URI
found && level==4 && current_name != "" && /^rdf:li[^=]+rdf:resource="/ {

  sub(/[^=]+resource="/,""); sub(/".*/,"")

  rdf_ctyp[++rdf_cnum] = "anyURI"
  rdf_ctyp[0]++

  rdf_cdat[rdf_cnum] = $0
  rdf_cdat[0]++

  # add URI-specific and misc. checks on rdf_cdat[rdf_cnum] here ...

}

# check that peer is declaring a supported CSA-RPC format.
found && level==2 && /^dc:format>/		{

  sub(/^dc:format>/,""); sub(/[^0-9.]/,""); rdf_fmt = $0
}

# detect </rdf:li>
found && level==4 && /^\/rdf:li>/		{next}

# detect </rdf:Seq>
found && level==4 && /^\/rdf:Seq>/ {current_name = "";level--; next}

# detect </param>
found && level==3 && /^\/csa:param>/		{level--; next}

# detect </result> and </error>
found && level==3 && /^\/csa:(result|error)>/{level--; next}

# detect </rdf:Description> (or whatever we called the outer element).
found && level==2 && $0 ~ rdf_element		{level--; next}

# end of RDF block.
found && level==1 && /^\/rdf:RDF>/		{level--; exit}

END {

  if (rc) exit(rc)

  if (!found || level) rc=150			# No RDF, or parse error.

  # Update this comparison as the CSA-RPC format evolves.
  if (rdf_fmt"" != "0.1") rc=151		 # Unknown format.

  # set default output file, just in case.
  if (ENVIRON["CSA_RPCC_AWK"] !~ /^\//)
		ENVIRON["CSA_RPCC_AWK"] = "/dev/stdout"

  print "BEGIN { _FORMAT=\"" _FORMAT "\" }" >> ENVIRON["CSA_RPCC_AWK"]

  _awkassign("_RDFCNAM", rdf_cnam, 0, 1, ENVIRON["CSA_RPCC_AWK"])
  _awkassign("_RDFCTYP", rdf_ctyp, 0, 1, ENVIRON["CSA_RPCC_AWK"])
  _awkassign("_RDFCDAT", rdf_cdat, 0, 1, ENVIRON["CSA_RPCC_AWK"])

  _awkassign("_RDFTYPE", rdf_type, 0, 1, ENVIRON["CSA_RPCC_AWK"])
  _awkassign("_RDFDATA", rdf_data, 0, 1, ENVIRON["CSA_RPCC_AWK"])
}

# =====================================================================
# End of csa-client.awk
# =====================================================================
