# =====================================================================
# AWK functions for CSA.
#
# Copyright (c) 2006,2007,2008,2009 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License 
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# string _mrwresponse(mrw [,string root, string child, string about,\
#	    string struct, string property, string name, string value,\
#	    array aname, array avalue, boolean relaxed, string file,\
#	    boolean noenc])
#
# Machine-readable Web (Semantic Web) response processor.
#
# This function provides CSA with the capability of generating
# responses that are readable according to the rules of the
# so-called Semantic Web. Here are some tips on how the
# underlying logic works, as it may look rather obscure ad a
# first glance. This program has proven flexible enough to handle
# the creation of rather complex and nested structures, like
# http://linuxgazette.net/105/misc/oregan/foaf-example.rdf.txt
#
# Here are the rules adopted, with respect to the above example:
#
# <foaf:Person> is a "root" element.
# <foaf:name>Jimmy O'Regan</foaf:name> is a name/value pair.
# <foaf:based_near> is a "child" element.
# <foaf:Point> is a "property" element.
# <geo:lat>52.6796</geo:lat> is a name/value pair.
#
# * Property elements can only appear inside RDF structs (Seq/Alt/Bag).
# * Root and property elements may take an "about" attribute, while 
#   child and name elements may not.
# * Property elements may only appear inside child nodes, which implies
#   that child nodes require an RDF struct.
# * Attributes are accepted only for name elements, and only if the
#   associated value is an URI.
# * Root elements may have name/value pairs, but not property elements
#   neither RDF structs, unless these are incapsulated inside child
#   elements.
# * A child element may not have name/value pairs, unless these are
#   incapsulated inside a struct of properties.
#
# Two optonal arrays of attribute names and corresponding values can
# be specified, but they are only taken into consideration for root
# elements and for name/value pairs in which "value" in an URI.
# 
# If the above rules are violated, the program will ignore the failing
# arguments and apply sensibly the correct ones, in order for the output
# to parse as valid XML/RDF. So if you wander why the output does not
# look like what you expected, either there's a bug in this program or
# you have got these rules wrong.
#
# This program looks flexible enough to allow for the the generation of
# arbitrarily complex XML structures, although I have not yet tested it
# extensively in that respect. If that were the case then it should be
# possible to produce entirely programmatic responses without requiring
# an external template, or requiring only a very minimal one.
#
# If the "file" argument is "!" then the result is returned to the
# caller as a (possibly huge) string, otherwise it is sent to the
# specified output stream.
#
# If option "relaxed" is set to "1" then the rules listed above become
# somewhat relaxed, as explained below, but the returned RDF may be
# less conformant:
#
# * Property elements do not need to be enclosed in RDF structs so
#   having them wrapped inside Seq/Alt/Bag becomes optional. Therefore,
#   property elements may appear inside child nodes without the need
#   for the child to contain an RDF struct.
# * Most (all?) elements take optional attributes, and even if the
#   associated value isn't a URI.
#
# The major limitation of this function, which was set by design, used
# to be the inability to geretare arbitrary depth XML structures. For
# instance, it was not possible to nest a child inside another child,
# or inside a property, with the function always closing the previous
# structure before starting the new one. With the introduction of
# multiple state arrays, the caller can now ask the function to generate
# the desired structures into separate arrays, and then embed one into
# anoher at will. This has proven to be extremely useful, and adds a
# great deal of power to this processor. Note that this will work even
# if the function is asked by the caller to print things to a file
# instead of returning them as a string. The caller will just have to
# get the unwanted values returned to it, and discard them, letting
# only the desired stuff go to the destination file.
# =====================================================================

function _mrwresponse(mrw,root,child,about,struct,property,name,value,\
		      aname,avalue,relaxed,file,noenc,	       a,s,S) {

  if (file == "") file = ENVIRON["CSA_RDF_MRW"]
  else if (file == "-") file = _STDOUT

  # make sure file is appended to, or printing multiple MRW
  # response blocks with different ontologies will not be possible!

  if (file != "!") printf(_NULL) >> file

  if (_isuri(tolower(about)) == _TRUE || about ~ /^#./)
		about = " rdf:about=\"" _xmlencode(about) "\""
  else if (about == ".") about = " rdf:about=\"\""
  else about = ""

  if (struct !~ /^rdf:/) struct = "rdf:" struct
  if (struct !~ /^rdf:(Alt|Bag|Seq)$/) {
     if (relaxed) struct = ""
     else struct = "rdf:Seq"				# strict default.
  }

  if (root != "") {					# new root ?

     # close previous tags if necessary.

     if (mrw["property"] != "") {

	s = sprintf("</%s>\n",mrw["property"])

	if (!relaxed) s = s sprintf("</rdf:li>\n")
	else if (struct != "") s = s sprintf("</rdf:li>\n")
     }

     if (mrw["struct"] != "") s = s sprintf("</%s>\n",mrw["struct"])

     if (mrw["child"] != "") s = s sprintf("</%s>\n",mrw["child"])

     if (mrw["root"] != "") s = s sprintf("</%s>\n",mrw["root"])

     if (s != "") {
	if (file == "!") S = S s
	else printf("%s",s) > file
     }

     # clear old values.
     mrw["child"] = mrw["struct"] = mrw["property"] = ""

     # the fake root value "." simply closes the MRW output.

     if (root == ".") {
	if (file != "!" && file != _STDOUT) close(file)
	mrw["root"] = _NULL		# this may be tested elsewhere.

	delete aname; delete avalue	# always do this before return!

	if (file == "!") {
	   sub(/\n$/,"",S)
	   return S
	}

	return
     }

     # set new root.
     s = sprintf("<%s%s",root,about)

     if (relaxed) {
     	for (a in aname)
	    s = s sprintf(" %s=\"%s\"",aname[a],_xmlencode(avalue[a]))
     }

     s = s ">\n"

     if (file == "!") S = S s
     else printf("%s",s) > file

     mrw["root"] = root

     # now that the array contents have been consumed I clear them,
     # as otherwise they would need to be cleared by the calling
     # program, or they could be attached also to other values they
     # do not apply to.

     delete aname; delete avalue
  }

  if (child != "") {					# new child ?

     # close previous tags if necessary.

     if (mrw["property"] != "") {

        s = sprintf("</%s>\n",mrw["property"])

	if (!relaxed) s = s sprintf("</rdf:li>\n")
	else if (struct != "") s = s sprintf("</rdf:li>\n")
     }

     if (mrw["struct"] != "") s = s sprintf("</%s>\n",mrw["struct"])

     if (mrw["child"] != "") s = s sprintf("</%s>\n",mrw["child"])

     if (s != "") {
	if (file == "!") S = S s
	else printf("%s",s) > file
     }

     # clear old values.
     mrw["struct"] = mrw["property"] = ""

     # the fake child value "." simply closes the current child.
     if (child == ".") {
	mrw["child"] = _NULL
	delete aname; delete avalue	# always do this before return!

	if (file == "!") {
	   sub(/\n$/,"",S)
	   return S
	}

	return
     }

     # set new child.
     s = sprintf("<%s",child)

     if (relaxed) {
	for (a in aname)
	    s = s sprintf(" %s=\"%s\"",aname[a],_xmlencode(avalue[a]))
     }

     if (!relaxed) s = s sprintf(">\n<%s",struct)
     else if (struct != "") s = s sprintf(">\n<%s",struct)

     s = s ">\n"

     if (file == "!") S = S s
     else printf("%s",s) > file

     mrw["child"] = child; mrw["struct"] = struct
  }

  if (property != "" && mrw["child"] != "") {		# new property ?

     # Properties can only occur inside childs (use default namespace).
     if (mrw["child"] == "") _mrwresponse(mrw,"","csa:Child")

     # close previous tags if necessary.

     if (mrw["property"] != "") {

	s = sprintf("</%s>\n",mrw["property"])

	if (!relaxed) s = sprintf("</rdf:li>\n")
	else if (struct != "") s = sprintf("</rdf:li>\n")

	if (file == "!") S = S s
	else printf("%s",s) > file
     }

     # the fake property value "." simply closes the current property.
     if (property == ".") {
	mrw["property"] = _NULL
	delete aname; delete avalue	# always do this before return!

	if (file == "!") {
	   sub(/\n$/,"",S)
	   return S
	}

	return
     }

     # set new property.

     if (!relaxed) s = sprintf("<rdf:li>\n")
     else if (struct != "") s = sprintf("<rdf:li>\n")

     s = s sprintf("<%s%s",property,about)

     if (relaxed) {
     	for (a in aname)
	    s = s sprintf(" %s=\"%s\"",aname[a],_xmlencode(avalue[a]))
     }

     s = s ">\n"

     if (file == "!") S = S s
     else printf("%s",s) > file

     mrw["property"] = property
  }

  if (name != "") {

     s = sprintf("<%s",name)

     if (!relaxed) {

	if (_isuri(tolower(value)) == _TRUE) {

	   for (a in aname)
	       s = s sprintf(" %s=\"%s\"",aname[a],_xmlencode(avalue[a]))

	   s = s sprintf(" rdf:resource=\"%s\" />\n",_xmlencode(value))
	}
	else {
	   if (noenc) s = s sprintf(">%s</%s>\n",value,name)
	   else s = s sprintf(">%s</%s>\n",_xmlencode(value),name)
	}
     }

     else {

	for (a in aname)
	    s = s sprintf(" %s=\"%s\"",aname[a],_xmlencode(avalue[a]))

	if (noenc) s = s sprintf(">%s</%s>\n",value,name)
	else s = s sprintf(">%s</%s>\n",_xmlencode(value),name)
     }

     # Turn into "<element [attrs] />" if no value was specified.
     if (value == "") sub (/><[^<]+>\n$/," />\n",s)

     if (s != "") {
	if (file == "!") S = S s
	else printf("%s",s) > file
     }
  }

  # now that the array contents have been consumed I clear them,
  # as otherwise they would need to be cleared by the calling
  # program, or they could be attached also to other values they
  # do not apply to.

  delete aname; delete avalue

  # Return resulting string, possibly huge.
  if (file == "!") {
     sub(/\n$/,"",S)
     return S
  }
}

